@@ -7,7 +7,7 @@ import pathspec
 
 from ..utils.fs import locate_file
 from .constants import DEFAULT_BUILD_DIRECTORY, BuildEnvVars
-from .utils import normalize_relative_directory, normalize_relative_path
+from .utils import normalize_inclusion_map, normalize_relative_directory, normalize_relative_path
 
 
 class BuilderConfig(object):
@@ -22,24 +22,31 @@ class BuilderConfig(object):
         self.__dependencies = None
         self.__sources = None
         self.__packages = None
+        self.__force_include = None
+        self.__vcs_exclusion_files = None
 
         # Possible pathspec.PathSpec
         self.__include_spec = None
         self.__exclude_spec = None
         self.__artifact_spec = None
-        self.build_artifact_spec = None
 
         # These are used to create the pathspecs and will never be `None` after the first match attempt
         self.__include_patterns = None
         self.__exclude_patterns = None
         self.__artifact_patterns = None
 
+        # Modified at build time
+        self.build_artifact_spec = None
+        self.build_force_include = {}
+
         # Common options
         self.__directory = None
+        self.__skip_excluded_dirs = None
         self.__ignore_vcs = None
         self.__only_packages = None
         self.__reproducible = None
         self.__dev_mode_dirs = None
+        self.__dev_mode_exact = None
         self.__require_runtime_dependencies = None
 
     @property
@@ -159,7 +166,7 @@ class BuilderConfig(object):
                 all_exclude_patterns.append(exclude_pattern)
 
             if not self.ignore_vcs:
-                all_exclude_patterns.extend(self.load_vcs_ignore_patterns())
+                all_exclude_patterns.extend(self.load_vcs_exclusion_patterns())
 
             if all_exclude_patterns:
                 self.__exclude_spec = pathspec.PathSpec.from_lines(
@@ -265,6 +272,26 @@ class BuilderConfig(object):
 
         return self.__directory
 
+    @property
+    def skip_excluded_dirs(self):
+        if self.__skip_excluded_dirs is None:
+            if 'skip-excluded-dirs' in self.target_config:
+                skip_excluded_dirs = self.target_config['skip-excluded-dirs']
+                if not isinstance(skip_excluded_dirs, bool):
+                    raise TypeError(
+                        'Field `tool.hatch.build.targets.{}.skip-excluded-dirs` must be a boolean'.format(
+                            self.plugin_name
+                        )
+                    )
+            else:
+                skip_excluded_dirs = self.build_config.get('skip-excluded-dirs', False)
+                if not isinstance(skip_excluded_dirs, bool):
+                    raise TypeError('Field `tool.hatch.build.skip-excluded-dirs` must be a boolean')
+
+            self.__skip_excluded_dirs = skip_excluded_dirs
+
+        return self.__skip_excluded_dirs
+
     @property
     def ignore_vcs(self):
         if self.__ignore_vcs is None:
@@ -377,6 +404,24 @@ class BuilderConfig(object):
 
         return self.__dev_mode_dirs
 
+    @property
+    def dev_mode_exact(self):
+        if self.__dev_mode_exact is None:
+            if 'dev-mode-exact' in self.target_config:
+                dev_mode_exact = self.target_config['dev-mode-exact']
+                if not isinstance(dev_mode_exact, bool):
+                    raise TypeError(
+                        'Field `tool.hatch.build.targets.{}.dev-mode-exact` must be a boolean'.format(self.plugin_name)
+                    )
+            else:
+                dev_mode_exact = self.build_config.get('dev-mode-exact', False)
+                if not isinstance(dev_mode_exact, bool):
+                    raise TypeError('Field `tool.hatch.build.dev-mode-exact` must be a boolean')
+
+            self.__dev_mode_exact = dev_mode_exact
+
+        return self.__dev_mode_exact
+
     @property
     def versions(self):
         if self.__versions is None:
@@ -563,6 +608,40 @@ class BuilderConfig(object):
 
         return self.__packages
 
+    @property
+    def force_include(self):
+        if self.__force_include is None:
+            if 'force-include' in self.target_config:
+                force_include_config = self.target_config
+                force_include_location = 'tool.hatch.build.targets.{}.force-include'.format(self.plugin_name)
+            else:
+                force_include_config = self.build_config
+                force_include_location = 'tool.hatch.build.force-include'
+
+            force_include = force_include_config.get('force-include', {})
+            if not isinstance(force_include, dict):
+                raise TypeError('Field `{}` must be a mapping'.format(force_include_location))
+
+            for i, (source, relative_path) in enumerate(force_include.items(), 1):
+                if not source:
+                    raise ValueError(
+                        'Source #{} in field `{}` cannot be an empty string'.format(i, force_include_location)
+                    )
+                elif not isinstance(relative_path, str):
+                    raise TypeError(
+                        'Path for source `{}` in field `{}` must be a string'.format(source, force_include_location)
+                    )
+                elif not relative_path:
+                    raise ValueError(
+                        'Path for source `{}` in field `{}` cannot be an empty string'.format(
+                            source, force_include_location
+                        )
+                    )
+
+            self.__force_include = normalize_inclusion_map(force_include, self.root)
+
+        return self.__force_include
+
     def get_distribution_path(self, relative_path):
         # src/foo/bar.py -> foo/bar.py
         for source, replacement in self.sources.items():
@@ -571,14 +650,47 @@ class BuilderConfig(object):
 
         return relative_path
 
-    def load_vcs_ignore_patterns(self):
-        # https://git-scm.com/docs/gitignore#_pattern_format
-        default_exclusion_file = locate_file(self.root, '.gitignore')
-        if default_exclusion_file is None:
-            return []
+    @property
+    def vcs_exclusion_files(self):
+        if self.__vcs_exclusion_files is None:
+            exclusion_files = {'git': [], 'hg': []}
 
-        with open(default_exclusion_file, 'r', encoding='utf-8') as f:
-            return f.readlines()
+            local_gitignore = locate_file(self.root, '.gitignore')
+            if local_gitignore is not None:
+                exclusion_files['git'].append(local_gitignore)
+
+            local_hgignore = locate_file(self.root, '.hgignore')
+            if local_hgignore is not None:
+                exclusion_files['hg'].append(local_hgignore)
+
+            self.__vcs_exclusion_files = exclusion_files
+
+        return self.__vcs_exclusion_files
+
+    def load_vcs_exclusion_patterns(self):
+        patterns = []
+
+        # https://git-scm.com/docs/gitignore#_pattern_format
+        for exclusion_file in self.vcs_exclusion_files['git']:
+            with open(exclusion_file, 'r', encoding='utf-8') as f:
+                patterns.extend(f.readlines())
+
+        # https://linux.die.net/man/5/hgignore
+        for exclusion_file in self.vcs_exclusion_files['hg']:
+            with open(exclusion_file, 'r', encoding='utf-8') as f:
+                glob_mode = False
+                for line in f:
+                    exact_line = line.strip()
+                    if exact_line == 'syntax: glob':
+                        glob_mode = True
+                        continue
+                    elif exact_line.startswith('syntax: '):
+                        glob_mode = False
+                        continue
+                    elif glob_mode:
+                        patterns.append(line)
+
+        return patterns
 
     def normalize_build_directory(self, build_directory):
         if not os.path.isabs(build_directory):
@@ -598,6 +710,11 @@ class BuilderConfig(object):
     def default_global_exclude(self):
         return ['.git', '__pycache__', '*.py[cod]']
 
+    def get_force_include(self):
+        force_include = self.force_include.copy()
+        force_include.update(self.build_force_include)
+        return force_include
+
     @contextmanager
     def set_build_data(self, build_data):
         try:
@@ -608,9 +725,12 @@ class BuilderConfig(object):
                     pathspec.patterns.GitWildMatchPattern, build_artifacts
                 )
 
+            self.build_force_include.update(normalize_inclusion_map(build_data['force_include'], self.root))
+
             yield
         finally:
             self.build_artifact_spec = None
+            self.build_force_include.clear()
 
 
 def env_var_enabled(env_var, default=False):
